/*



 Realisiert mit ATmega88 an 4MHz Oszillator




             Reset PC6 1  28 PC5   zum Transistor
                Rx PD0 2  27 PC4   RxOut (an Pin2)
                   PD1 3  26 PC3   Sekundentakt (Benutzung als volatiles Bit)
             RS232 PD2 4  25 PC2   Blinkt im Sekundentakt
             RS232 PD3 5  24 PC1   
             IstAn PD4 6  23 PC0   
                   VCC 7  22 GND
                   GND 8  21 AREF
               Q1  PB6 9  20 AVCC
               Q2  PB7 10 19 PB5   SCK   
                   PD5 11 18 PB4   MISO  
                   PD6 12 17 PB3   MOSI  
                   PD7 13 16 PB2   
                   PB0 14 15 PB1   

  RXD: Pin 2  PD0
  RxOut: Pin27 PC4   verbunden mit Pin2
  
  RS232-Daten auf diese beiden Eingnge:
  INT0: Pin 4 PD2  Verbunden mit
  INT1: Pin 5 PD3  (Int0 macht fallende, INT1 steigende Flanke)

  IstAn: Pin 6 PD4  Rechner ist an

  PC5: Pin28  Ausgang zum Transistor
  PC3: Pin26  Sekundentakt


  Das Programm empfngt via RS232 eine Zahl in Dezimalschreibweise mit 
  vorangestelltem '*' und abschlieendem CR (#13). Also z.B. *86400
  Die Zahl wird runtergezhlt und dann fr 1 Sekunde der Ausgangstransistor eingeschaltet,
  sofern Pin 6 LOW ist.

  Besonderheiten:
  Statt Max232 wird einfach ein 47K Widerstand und eine 5V1 Z-Diode genommen. 
  (Die Schutzdioden halten bis 1mA)
  Zur Signalinvertierung dienen die Eingnge Pin4+5. 
  Das invertierte Signal steht dann an Pin27 an und geht an den RX Eingang (Pin2)

  Da der Compiler Probleme mit volatile-Variablen hat, wird zum Info-Transfer zwischen 
  Interruptroutine und Hauptprogramm ein Portpin benutzt (PC3).

  

*/

#define F_CPU 4000000UL	/* CPU clock in Hertz */
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>


#define U8 unsigned char
#define U16 unsigned int
#define U32 unsigned long
#define I16 int

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128

U8 rxbuffer[20];
U8 rxIndexStore=0;
U8 rxIndexRead=0;
//--------------------------------------------------------
void __attribute__ ((naked)) INT0_vect(void){ 
//  asm("sbi	0x18, 4");   
  PORTC|=BIT4;  // High ausgeben
  asm("reti");
}
//--------------------------------------------------------
void __attribute__ ((naked)) INT1_vect(void){ 
  PORTC&=~BIT4; // Low ausgeben
  asm("reti");
}


//--------------------------------------------------------
// RS232 Empfang im Hintergrund in FIFO
ISR(USART_RX_vect){
  rxbuffer[rxIndexStore++] = UDR0;
  if (rxIndexStore>19) rxIndexStore=0;
}
//--------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  PORTC|=BIT3;  // High ausgeben
}
//--------------------------------------------------------
void USART_Init(){
//--------------------------------------------------------

  /* Set baud rate: round(f_osc / (16*baudrate)) - 1  wenn U2X=0*/
  UBRR0H = 0;
  UBRR0L = 25; //9600 Baud bei 4MHz

  UCSR0A=0; //Clear U2X

  /* Set frame format:          2stop bit   8data,*/
  UCSR0C =  (1<<USBS0)|(3<<UCSZ00);

  // RX enable und IRQ auf RX
  UCSR0B = (1<<RXCIE0) | (1<<RXEN0);
}
//--------------------------------------------------------
#define rxDataAvail (rxIndexStore!=rxIndexRead)
//char rxDataAvail(void){
//  return (rxIndexStore<>rxIndexRead);
//}
//--------------------------------------------------------
char getRxChar(void){
  static char zeichen;
  zeichen = rxbuffer[rxIndexRead++];
  if (rxIndexRead>19) rxIndexRead=0;
  return zeichen;  
}

//--------------------------------------------------------
U32 receiveLine(void){
  static U8 step=0;
  static U32 zahl=0;
  U8 zeichen;
  while (rxDataAvail) {
    zeichen = getRxChar();
    switch(step){
      case 0:   //warte auf Startzeichen
        zahl=0;
        if ('+'==zeichen) step=1; 
        break;
      case 1:
        if ((zeichen>='0') && (zeichen <='9')){
          zahl = 10 * zahl + zeichen - '0';
          if (zahl > 32000000L){ //ber ein Jahr
            step=0;
          }
        }else{
          if ('+'==zeichen){
            zahl=0;
          }else{
            step=0;
            if ('s'==zeichen){ //Zeilenendekennung
              return zahl+1;
            }
          }
        }
        break;
      default: step=0; break;  
    }//switch  
  }//while
  return 0; //Keine komplette Zahl
}
//--------------------------------------------------------

int main(void){
U32 sekundenBisStart=7;
U32 empfangeneZahl=0;
U8 n;
  cli();
  DDRC=           BIT2|BIT3|BIT4|BIT5;
  PORTB=0xff;
  PORTC=BIT0|BIT1; //Pullup an unbenutzten Eingngen
  PORTD=     BIT1|BIT2|BIT3|     BIT5|BIT6|BIT7;

 
  EICRA = (1<<ISC01) | (1<<ISC11)|(1<<ISC10); //falling edge INT0, rising edge INT1 generates an interrupt request.
  EIMSK = (1<<INT0)|(1<<INT1); //Enable INT0+INT1

  ACSR = (1<<ACD); // Disable comparator
  ADCSRA=0;        // Disable A/D converter

  USART_Init(); // setup the UART
 
  //Timer 1 erzeugt Sekundentakt
  //Set up the 16-bit timer 1.
  TCCR1A =  0; //Normaler Timer/Counter
  TCCR1B =  _BV(WGM12) | _BV(CS12) ; // CTC-Betrieb, Vorteiler 256  -> 15625Hz
  OCR1A = 15624;  //Einen weniger als 15625
  TIMSK1 = _BV(OCIE1A); 	// Timer-Interrupt aktivieren
  //TIFR wird bei Interrupt gesetzt

  sei();

  while(1){
    if (  PORTC & BIT3){ // Sekunde rum
      PORTC &=~BIT3; //Bit zurcksetzen
      if (PORTC & BIT2) PORTC &=~BIT2; else PORTC |=BIT2; //LED Blinken
      if (sekundenBisStart >0){
        sekundenBisStart--;
        if (1==sekundenBisStart){
          if (0==(PIND & BIT4)){ //Nur einschalten, wenn Rechner aus ist
            PORTC |= BIT5; //Einschalten 
            _delay_ms(800);
          }else{
            //LED Blinken zum debuggen
            for (n=0; n<100; n++){ 
              PORTC &=~BIT2; 
              _delay_ms(40);
              PORTC |=BIT2; 
              _delay_ms(40);
            }
          }
        }
      }
    }//Sekunde rum

    PORTC &=~BIT5; //nicht Einschalten

    empfangeneZahl = receiveLine();
    if ( empfangeneZahl > 0 ){
      sekundenBisStart = empfangeneZahl;
      //LED Blinken zum debuggen
      for (n=0; n<6; n++){ 
        PORTC &=~BIT2; 
        _delay_ms(40);
        PORTC |=BIT2; 
        _delay_ms(40);
      }
    };
  }//while 1
}
